
/* remap packbuilder enum's to defines */
#define ADDING_OBJECTS   GIT_PACKBUILDER_ADDING_OBJECTS
#define DELTAFICATION    GIT_PACKBUILDER_DELTAFICATION

#include "const-c-packbuilder.inc"

#undef ADDING_OBJECTS
#undef DELTAFICATION

