
/* remap libgit2 class enum's to defines */
#define NONE          GITERR_NONE
#define NOMEMORY      GITERR_NOMEMORY
#define OS            GITERR_OS
#define INVALID       GITERR_INVALID
#define REFERENCE     GITERR_REFERENCE
#define ZLIB          GITERR_ZLIB
#define REPOSITORY    GITERR_REPOSITORY
#define CONFIG        GITERR_CONFIG
#define REGEX         GITERR_REGEX
#define ODB           GITERR_ODB
#define INDEX         GITERR_INDEX
#define OBJECT        GITERR_OBJECT
#define NET           GITERR_NET
#define TAG           GITERR_TAG
#define TREE          GITERR_TREE
#define INDEXER       GITERR_INDEXER
#define SSL           GITERR_SSL
#define SUBMODULE     GITERR_SUBMODULE
#define THREAD        GITERR_THREAD
#define STASH         GITERR_STASH
#define CHECKOUT      GITERR_CHECKOUT
#define FETCHHEAD     GITERR_FETCHHEAD
#define MERGE         GITERR_MERGE
#define SSH           GITERR_SSH
#define FILTER        GITERR_FILTER
#define REVERT        GITERR_REVERT
#define CALLBACK      GITERR_CALLBACK
#define CHERRYPICK    GITERR_CHERRYPICK
#define DESCRIBE      GITERR_DESCRIBE
#define REBASE        GITERR_REBASE
#define FILESYSTEM    GITERR_FILESYSTEM

#include "const-c-category.inc"

#undef NONE
#undef NOMEMORY
#undef OS
#undef INVALID
#undef REFERENCE
#undef ZLIB
#undef REPOSITORY
#undef CONFIG
#undef REGEX
#undef ODB
#undef INDEX
#undef OBJECT
#undef NET
#undef TAG
#undef TREE
#undef INDEXER
#undef SSL
#undef SUBMODULE
#undef THREAD
#undef STASH
#undef CHECKOUT
#undef FETCHHEAD
#undef MERGE
#undef SSH
#undef FILTER
#undef REVERT
#undef CALLBACK
#undef CHERRYPICK
#undef DESCRIBE
#undef REBASE
#undef FILESYSTEM

